# -*- coding: utf-8 -*-
#######################################################################

###        ABACOM netPIO /freq board (Python 2 script)
###        
###      tested on Raspberry Pi Rev.2 (Linux)  ABACOM 11/2013 

###    other linux systems were not tested. Feedback wellcome.

#######################################################################

#######################################################################
### usage: netpio_freq.py -a <host or IP> -p <UDP port> [-r]
#######################################################################

### example calls from linux command line:

###     python ./netpio_freq.py --address 192.168.1.xxx --port 37157  [-r]
### or
###     python ./netpio_freq.py -a myhostname.somewhere.com -p 37157 [-r]

### option: [-r] for READ-ONLY request

###################### required libraries  #######################

### These libs likely are already pre-installed on RaspPi...
import sys
import time
import getopt
import socket

# Wrap the (USB) request into UDP datagramm
# see netPIO datasheet http://www.abacom-online.de/div/FirmwareDocs/UK/ABACOM_netPIO.pdf
def wrapRequest(aRequest, responseLen, syncByte):
    # add leading "TXDATA" 
    aRequest.insert(0, "T")
    aRequest.insert(1, "X")
    aRequest.insert(2, "D")
    aRequest.insert(3, "A")
    aRequest.insert(4, "T")
    aRequest.insert(5, "A")
    # The USB request is this position
    # add length we exspect (or use syncbyte)
    aRequest.append(responseLen)
    aRequest.append(syncByte)

def buildGetRequest():
    request = bytearray() # freq request is 10 bytes long in any case
    # byte 1..9 not used in GET request
    request.append(0)
    request.append(0)
    request.append(0)
    request.append(0)
    request.append(0)
    request.append(0)
    request.append(0)
    request.append(0)
    request.append(0)
    # byte 10: set Bit 7 to indicate a GET request (not in USB version)
    request.append(0x80) 
    return request

def buildSetRequest():
    request = bytearray() # freq request is 10 bytes long in any case

    ### Byte 1 - Gate & DOUT
    # Bits 37 are assigned to the digital outputs B2, B3, B5, D3 and D6, one after each other. 
    # These bits control the output state and activate alternative functions.
    B2on = 0x08 
    B3on = 0x10
    B5on = 0x20
    D3on = 0x40
    D6on = 0x80
    GEN1on = D6on # Generator 1 is output @D6  (8 bit timer/counter)
    GEN2on = D3on # Generator 2 is output @D3  (16 bit timer/counter)
    GEN3on = B3on # Generator 3 is output @B3  (PWM 72kHz, 8 bit)
    # Bits 0..2 select the gate time interval for the frequency measurement
    Gate2000ms = 0x00
    Gate1000ms = 0x01
    Gate500ms  = 0x02
    Gate100ms  = 0x03
    Gate50ms   = 0x04
    Gate20ms   = 0x05
    Gate10ms   = 0x06
    GateOff    = 0x07 # Gate off = free running counter
    ### We use Gate time 1s turn D3 on for GEN2 output
    request.append(Gate1000ms + GEN2on) 

    ### Byte 2 - GEN1 compare value
    # This byte adjust the compare value of generator 1 and varies output frequency 
    # or PWM depending on the generator mode.
    ### we set to middle position here...
    request.append(0x7F) 

    ### Byte 3 - GEN3 compare value
    # This byte adjust the compare value of generator 3 and varies PWM.
    ### We set PWM to 50% here...
    request.append(0x7F) 

    ### Bytes 4 - GEN1 Mode/Source , GEN2 timebase
    # Bits 02 select the clock source for generator 1.
    Gen1ClkOff          = 0x00 # Clock off
    Gen1Clk18MHZ432     = 0x01 # 18.432 MHz
    Gen1Clk2MHZ304      = 0x02 #  2.304 MHz
    Gen1Clk288KHZ       = 0x03 # 288 kHz
    Gen1Clk72KHZ        = 0x04 #  72 kHz
    Gen1Clk18KHZ        = 0x05 #  18 kHz
    Gen1ClkExtFalling   = 0x06 # External Clock (D4 falling)
    Gen1ClkExtRising    = 0x07 # External Clock (D4 rising)
    # Bit 3 resets the (frequency) counter to zero (Counter Software Reset), if bit is SET.
    CounterNoReset      = 0x00  
    CounterReset        = 0x08  
    # Bit 4 and Bit 5 select the mode of generator 1
    Gen1ModeNormal      = 0x00 # Fout = ( Fin / 2 ) / 256
    Gen1ModePWM         = 0x10 # Fout = ( Fin / 2 ) / 255
    Gen1ModeCTC         = 0x20 # Fout = ( Fin / 2 ) / ( CMP1 + 1 )
    Gen1ModeFastPWM     = 0x30 # Fout = ( Fin / 256 )
    # Bit 6 and Bit 7 select the timebase for generator 2.
    GEN2Timebase125us   = 0x00 # 125 microseconds
    GEN2Timebase500us   = 0x40 # 500 microseconds
    GEN2Timebase5ms     = 0x80 # 5 milliseconds
    GEN2Timebase500ms   = 0xC0 # 500 milliseconds
    ### We do not use GEN1 and set GEN2timebase to 0.125 ms here...
    # No need to reset the counter (we already set gate time to 1s)
    request.append(Gen1ClkOff + Gen1ModeNormal + GEN2Timebase125us + CounterNoReset) 

    ### Generator 2 - we like to generate a 1kHz test signal @D3 ...
    ### for a simple test you can connect (D3 out) -> (D2 in) and (D5 in)
    ### this should give you a 1000 Hz reading

    ### Bytes 5,6 (16 bit word) - GEN2 Duration high phase(T high = Duration * GEN2timbase)
    request.append(0x04) # LowByte 4*0,125ms = 0,5 ms
    request.append(0x00) # HighByte
    ### Bytes 7,8 (16 bit word) - GEN2 Duration low phase( T low  = Duration * GEN2timbase)
    request.append(0x34) # LowByte 4*0,125ms = 0,5 ms
    request.append(0x00) # HighByte

    ### Byte 9 Configuration And Digital Out (Flags)
    B0resetsCounter    = 0x01 # Bit 0: B0 becomes RESET input for the (frequency) counter.
    B1triggersGen2     = 0x02 # Bit 1: B1 becomes SYNC/TRIGGER input for generator 2.
    B4enablesCounter   = 0x04 # Bit 2: B4 becomes ENABLE input for the (frequency) counter.
    GEN2singleshotMode = 0x08 # Bit 3: sets generator 2 mode to single-shot
    GEN2continuousMode = 0x00 # 
    GEN2softtrigger    = 0x10 # Bit 4 is SYNC/TRIGGER for generator 2 (software trigger)	    
    D4enablesGEN2clock = 0x20 # Bit 5: Use D4 as CLOCK ENABLE input for generator 2
    D7outputsGateClock = 0x40 # Bit 6: activates Gate Clock Output at D7.
    D7on 	       = 0x80 # Bit 7: turns output D7 on
    ### We use GEN2 in continious mode, output gate clock at D7 and turn it on ...
    request.append(GEN2continuousMode + D7outputsGateClock + D7on) 

    ### Byte 10 VREF select 
    VrefExternal       = 0x00 # External from VREF terminal
    VrefInternalVcc    = 0x01 # Internal (Vcc) approx. 5V
    VrefInternal1V1    = 0x03 # Internal 1.1 Volt
    ### We prefer VREF = Vcc here...
    request.append(VrefInternalVcc) 

    return request

# dynamic rounding of freq due to gate time
# to get rid of noisy digits
def roundFreq(aFreq, aGate):
    result=aFreq
    digits=7 # default
    if aFreq>=0.1:
        if aGate==255: 
            digits=5
        elif aGate==127:
            digits=5
        elif aGate==63: 
            digits=5
        elif aGate==31: 
            digits=5
        elif aGate==15: 
            digits=5
        elif aGate==7: 
            digits=5
        elif aGate==3: 
            digits=5
        elif aGate==0: 
            digits=-1
    if aFreq>=1:
        if aGate==255:
            digits=4
        elif aGate==127:
            digits=4
        elif aGate==63: 
            digits=4
        elif aGate==31: 
             digits=4
        elif aGate==15: 
            digits=4
        elif aGate==7: 
            digits=4
        elif aGate==3: 
            digits=4
        elif aGate==0: 
            digits=-1
    if aFreq>=10:
        if aGate==255: 
            digits=3
        elif aGate==127: 
            digits=3
        elif aGate==63: 
            digits=3
        elif aGate==31: 
            digits=3
        elif aGate==15:
            digits=3
        elif aGate==7: 
            digits=3
        elif aGate==3:
            digits=3
        elif aGate==0:
            digits=-1
    if aFreq>=100:
        if aGate==255:
            digits=3
        elif aGate==127: 
            digits=3
        elif aGate==63: 
            digits=3
        elif aGate==31:
            digits=2
        elif aGate==15: 
            digits=2
        elif aGate==7: 
            digits=2
        elif aGate==3:
            digits=2
        elif aGate==0: 
            digits=-1
    if aFreq>=1000:
        if aGate==255:
            digits=4
        elif aGate==127: 
            digits=3
        elif aGate==63: 
            digits=3
        elif aGate==31: 
            digits=2
        elif aGate==15: 
            digits=2
        elif aGate==7: 
            digits=2
        elif aGate==3: 
            digits=2
        elif aGate==0: 
            digits=-1
    if aFreq>=10000:
        if aGate==255:
            digits=5
        elif aGate==127: 
            digits=4
        elif aGate==63: 
            digits=4
        elif aGate==31: 
            digits=4
        elif aGate==15: 
            digits=3
        elif aGate==7: 
            digits=3
        elif aGate==3: 
            digits=3
        elif aGate==0: 
            digits=-1
    if aFreq>=100000:
        if aGate==255:
            digits=6
        elif aGate==127: 
            digits=5
        elif aGate==63: 
            digits=5
        elif aGate==31: 
            digits=4
        elif aGate==15: 
            digits=4
        elif aGate==7: 
            digits=3
        elif aGate==3: 
            digits=3
        elif aGate==0: 
            digits=-1
    if aFreq>=1000000:
        if aGate==255:
            digits=7
        elif aGate==127: 
            digits=5
        elif aGate==63: 
            digits=4
        elif aGate==31:
            digits=4
        elif aGate==15: 
            digits=4
        elif aGate==7: 
            digits=3
        elif aGate==3: 
            digits=3
        elif aGate==0: 
            digits=-1
    if digits>0:
        result=roundToDigits(aFreq,digits)
    return result

# round to a given nuber of digits (not decimals)
# roundToDigits(1234.56789 , 7) = 1234.567
# roundToDigits(1234.56789 , 5) = 1234.5
# roundToDigits(1234.56789 , 2) = 1200.0
def roundToDigits(aFloat,digits):
    c=0
    while aFloat>1:
        aFloat=1.0*aFloat/10
        c=c+1
    aFormat = "%."+str(digits)+"f"
    r = aFormat % aFloat
    aFloat = float(r)
    while c>0:
        aFloat=1.0*aFloat*10
        c=c-1
    return aFloat
    
# execute a netPIO /freq command string via UDP
def netpiofreqrequest(aHost, aPort, request):

    # print the request bytes to debug
    # print "Request = ", str.join(" ", ("%02x" % i for i in request)) 

    response = bytearray()

    ### build the request...

    freqResponseLen = 33 # freq response is 33  bytes long
    UnusedSyncByte = 0   # we do not use a sync byte
    
    if len(request)<>10: # check the length, you in case...
        print "Bad request length, must be 10!"
        sys.exit

    # Wrap the request into UDP datagramm 
    wrapRequest(request, freqResponseLen, UnusedSyncByte)

    ### ...send the request...

    # get a UDP socket
    UDP = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # (Internet, UDP)
    # and fire datagramm on UDP
    UDP.sendto(request, (aHost, aPort))


    ### ... and look for the response...

    UDP.settimeout(2.00) # float timeout value in seconds
    try:
        # no need to bind, we are bound by previous send 
        # sock.bind((UDP_IP, UDP_PORT))
        response, addr = UDP.recvfrom(freqResponseLen) # receive from UDP
        
    except socket.timeout:
        print "No response from ",aHost," (",aPort,")"
        sys.exit()
    return response

# decode and printout the response  
def niceList(data):

    # data string to bytearray()
    response = bytearray(b"")
    for c in data: response.append(c)
     
    # print the response bytes to debug
    # print "Response = ", str.join(" ", ("%02x" % i for i in response)) 
 
    # response array is zero-based, but documentation is 1-based
    # so we match response[index] with documentation for better readability...
    response.insert(0,0x00) # we insert a dummy byte response[0] 
    # now we can decode easily...

    ##### Bytes 1...12: edge measurement result 
    # 3x 32 bit time samples from free-running counter, clocked with 720kHz
    # D2 input triggers time sampling
    print "Edge measurement result (D2 input):"
    # time stamp of previous edge 
    T1 = (response[4] << 24) + (response[3] << 16) + (response[2] << 8) + response[1]
    # time stamp of latest edge 
    T2 = (response[8] << 24) + (response[7] << 16) + (response[6] << 8) + response[5]
    # current time stamp 
    T3 = (response[12]<< 24) + (response[11]<< 16) + (response[10] << 8) + response[9]
    # We calculate (latest - previous edge) / 72 kHz = cycle time in seconds
    T = 1.0/72000 * (T2-T1) 
    if T>0: 
        # frequency = 1 / cycle time
        F1 = 1 / T # Frequency from D2 (0..approx. 200Hz)
        print " Cycle time T  =",T, "s"
        print " Frequency 1   =",F1, "Hz" # we found our first (low) frequency value!
    else: # no signal or frequency too high
        F1 = 0 # we do not have time samples and can not calc a freq
        print " invalid"
        print " invalid"
    # We calculate (current time - latest edge) / 72 kHz = time ellapsed since last edge in seconds
    if T3>=T2:
        TimeEllapsed = 1.0/72000 * (T3-T2) 
    else:
        TimeEllapsed = 0
    print " Time Ellapsed =",TimeEllapsed, "s"
    ####

    #### Bytes 13...16 counter value
    # Frequency / Counter / Gate measurement
    print "Counter read result (D5 input):"
    CNT = (response[16] << 24) + (response[15] << 16) + (response[14] << 8) + response[13]
    print " Count =",CNT # we get (gated) the counter value
    # Byte 20: Gate time
    GatetimeMS = response[20] *10 # Gatetime in milliseconds
    if GatetimeMS>0:
       print " Gate  =",GatetimeMS, 'ms'
       F2 = 1.0E3 * CNT / GatetimeMS # Frequency = Counts per Gatetime 
       print " Freq  =", F2, "Hz" # we found a second (high) frequency
    else:
       print " Gate = OFF"

    # Byte 17,18,19: Digital ports 
    print "Digital port status:76543210"
    print " Port B  =", "0x%02x" % response[17],' = ',bin(response[17])[2:].zfill(8)
    print " Port C  =", "0x%02x" % response[18],' = ',bin(response[18])[2:].zfill(8)
    print " Port D  =", "0x%02x" % response[19],' = ',bin(response[19])[2:].zfill(8)
    
    # Byte (21,22) to (31,32): ADC raw values
    print "ADC read results:"
    for i in range(0,6):
        ADCraw = (response[22 + (i*2) ] << 8) + response[21 + (i*2)] # raw adc value 0..1023
        ADChex = "0x%03x" % ADCraw # to HEX string
        ADCpercent = "%0.1f" % (1.0 * ADCraw/1023*100) # to percent string 
        print " ADC"+str(i)+": ", ADChex," = ", ADCpercent+"% * Vref"

    # Byte 33 reserved
    
    # We combine both measurements for best result
    if (T>0):
       BestFreq = F1     # we have a result from edge measurement (low freqs)
    else:
       BestFreq = F2     # otherwise use result from freq counter (high freqs)

    # BestFreq - that is our numeric frequency value
    
    # we can cut-off noisy digits that are out of measurable precision
    Gate = response[20]
    BestFreq = roundFreq( BestFreq , Gate )  # dynamic noise cut-off depending on gate time
    
    # finally you may prefer friendly display format with simple auto-range
    aFormat = " f = %.5f Hz"
    if BestFreq>=1:
        aFormat = " f = %.4f Hz"
    if BestFreq>=10:
        aFormat = " f = %.3f Hz"
    if BestFreq>=100:
        aFormat = " f = %.2f Hz"
    if BestFreq>=1000:
        BestFreq = BestFreq / 1000
        aFormat = " f = %.4f kHz"
        if BestFreq>=10:
            aFormat = " f = %.3f kHz"
        if BestFreq>=100:
            aFormat = " f = %.2f kHz"
        if BestFreq>=1000:
            BestFreq = BestFreq / 1000
            aFormat = " f = %.4f MHz"
    print "friendly result:"
    print  aFormat %  BestFreq
     
      

### main method processes the command line ...
def main(argv):

    ### Get string args from command line...
    udpHostString= ''
    udpPortString = ''
    performSetRequest = True

    try:
        opts, args = getopt.getopt(argv,"hra:p:",["address=","port="])
    except getopt.GetoptError:
        print 'usage: sudo netpio_freq.py -a <host> -p <port> [-r]'
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print 'usage: sudo netpio_freq.py -a <host> -p <port> [-r]'
            sys.exit()
        elif opt in ("-a", "--address"):
            udpHostString = arg
        elif opt in ("-p", "--port"):
            udpPortString = arg
        elif opt == '-r':
            # we only want read, without changing any settings
            performSetRequest = False 
    ### We got the args now, so we can try to use it....

    ### Check UDP host from command line...
    if udpHostString=='': # no host specified in command line
        print 'No host specified!' 
        print 'usage: sudo netpio_freq.py -a <host> -p <port> [-r]'
        sys.exit()

    ### Now get UDP port from command line...
    if udpPortString=='': # no port specified in command line
        print 'No port specified, using default port 37155.' 
        udpPort = 37155
    else:
        try:
            udpPort = int(udpPortString)
        except ValueError:
            print 'Invalid port value!'
            sys.exit()

    if not (udpPort in range(1,65534)):
        print 'Port value out of range (must be 1..65534)'
        sys.exit()

    ### Hurray! We have host and port value.
    ### Here we go...
    # see freq firmware docs for protokol details
    # http://www.abacom-online.de/div/FirmwareDocs/UK/Firmware_freq.pdf

    if performSetRequest:
       # we build new settings
       some_request=buildSetRequest() 
       # and send it to device
       netpiofreqrequest(udpHostString, udpPort, some_request) 
       # wait a some time until new settings settled
       time.sleep(2.00) 
       # build a read only request  
       some_request=buildGetRequest() 
       # send the read request and printout the response
       niceList( netpiofreqrequest(udpHostString, udpPort, some_request) ) # Printout the response
    else:
       # we only read (-r in params) but do not change the settings
       some_request=buildGetRequest()
       niceList( netpiofreqrequest(udpHostString, udpPort, some_request) ) # Printout the response

    ### finish!  
          
##############################################################
### call the main method with arguments from command line ...
if __name__ == "__main__":
   main(sys.argv[1:])
